require "ruuuby/extensions/version"

# This module + gem simply aims to provide coding quality of life extensions to core classes, somewhat similar to Python's approach.
module Ruuuby

  # Empty module. To be used for documenting Ruuuby.
  module Extensions
    # Placeholder error class for errors specifically caused by Ruuuby.
    class Error < StandardError; end
  end

  # Utility functionalities for checking data types and throw exceptions.
  module Params
    # @param [String|*] str
    #
    # @return [Boolean] true, if provided param is not nil and inherits from String class.
    def self.not_string?(str)
      str == nil || !str.is_a?(String)
    end

    # @param [Array|*] str
    #
    # @return [Boolean] true, if provided param is not nil and inherits from Array class.
    def self.not_array?(ary)
      ary == nil || !ary.is_a?(Array)
    end

    # @param [String|nil|*] param
    #
    # @raise [ArgumentError]
    #
    # @return [Boolean] true only, exception raised otherwise
    def self.check_string(param)
      raise ArgumentError.new(self::bad_type(caller_locations.first.label.to_s, String, param)) if Ruuuby::Params::not_string?(param)
      true
    end

    # @param [Array|nil|*] param
    #
    # @raise [ArgumentError]
    #
    # @return [Boolean] true only, exception raised otherwise
    def self.check_array(param)
      raise ArgumentError.new(self::bad_type(caller_locations.first.label.to_s, Array, param)) if Ruuuby::Params::not_array?(param)
      true
    end

    # @param [Array|nil|*] param
    #
    # @raise [ArgumentError]
    #
    # @return [Boolean] true only, exception raised otherwise
    def self.check_char(param)
      if param == nil || !param.is_a?(String)
        error_message = "a param of type{#{String.to_s}}, instead got{#{param.class.to_s}}!"
      elsif param.chars.length != 1
        error_message = "a String of length 1, instead got{#{param.length.to_s}}"
      else
        return true
      end
      raise ArgumentError.new("#{caller_locations.first.label.to_s} requires #{error_message}")
    end

    private

    # @param [String] func | the name of the original function that triggered this exception
    # @param [*] required_type | the required data type for the param
    # @param [*] param | the variable with mis-mitached type
    #
    # @return [String]
    def self.bad_type(func_name, required_type, param)
      "#{func_name} requires param of type{#{required_type.to_s}}, instead got{#{param.class.to_s}}!"
    end
  end
end

unless Array.respond_to?(:equal_contents?)

  # Add extensions to the default class(*String*).
  class Array

    # @param [Array] them | another array to be compared against
    #
    # @raise [ArgumentError]
    #
    # @return [Array] a new array instance containing values only appearing in base array (and not in provided array)
    def uniq_to_me(them)
      Ruuuby::Params::check_array(them)
      self.uniq - them.uniq
    end

    # @param [Array] them | an array to be compared against
    #
    # @raise [ArgumentError]
    #
    # @return [Array] a new array instance containing values only appearing in provided array (and not in base array)
    def uniq_to_them(them)
      Ruuuby::Params::check_array(them)
      them.uniq - self.uniq
    end

    # @param [Array] them | an array to be compared against
    #
    # @raise [ArgumentError]
    #
    # @return [Array] a new array instance containing values only appearing in provided array (and not in base array)
    def shared_between(them)
      Ruuuby::Params::check_array(them)
      self.uniq & them.uniq
    end

    # @return [Array]
    def remove_empty!
      return self if self.length == 0
      compact!.reject!{|element| element.respond_to?('empty?') && element.empty?}
    end

    # @param [Array] them
    #
    # @raise [ArgumentError]
    #
    # @return [Boolean] true, if this array and provided array both have same content (*order-matters*: false, *duplicates-matter*: false)
    def equal_contents?(them)
      Ruuuby::Params::check_array(them)
      return true  if (self.length == 0 && them.length == 0)
      return false if (self.length != them.length)
      uniq_to_me(them).empty? && uniq_to_them(them).empty?
    end

    # @param [Array] ending
    #
    # @raise [ArgumentError]
    #
    # @return [Boolean] true, if this array ends with the elements in the provided parameter array (*order-matters*: true, *duplicates-matter*: true)
    def end_with?(ending)
      Ruuuby::Params::check_array(ending)
      if ending.length == 0
        self.length == 0
      else
        self[(self.length-ending.length)..(self.length-1)] == ending[(0)..(ending.length-1)]
      end
    end

    # @param [Array] ending
    #
    # @raise [ArgumentError]
    #
    # @return [Array]
    def ensure_ending!(ending)
      Ruuuby::Params::check_array(ending)
      len_ending = ending.length
      return self if len_ending == 0 || self.end_with?(ending)
      len_this = self.length
      return self << ending[0] if len_ending == 1
      if len_this == 0
        ending.each { |element| self << element }
        return self
      end
      delta        = 0
      last_matched = nil
      while delta <= len_this && delta <= len_ending
        ending_of_this  = self[(len_this-1-delta)..(len_this-1)]
        starting_of_end = ending[0..delta]
        last_matched    = starting_of_end if ending_of_this == starting_of_end
        delta          += 1
      end
      if last_matched == nil
        ending.each { |element| self << element }
      else
        ending[(last_matched.length)..(len_ending-1)].each { |element| self << element }
      end
      self
    end
  end

  # Add extensions to the default class(*String*).
  class String

    # @param [String] ending
    #
    # @raise [ArgumentError]
    #
    # @return [String]
    def ensure_ending!(ending)
      Ruuuby::Params::check_string(ending)
      return self if ending == '' || self.end_with?(ending)
      len_this = self.length
      return self << ending if len_this == 0
      last_matched = ''
      len_ending   = ending.length
      delta        = 0
      while delta <= len_this && delta <= len_ending
        ending_of_this  = self[(len_this-1-delta)..(len_this-1)]
        starting_of_end = ending[0..delta]
        last_matched    = starting_of_end if ending_of_this == starting_of_end
        delta          += 1
      end
      self << (last_matched == '' ? ending : ending[last_matched.length..len_ending-1])
    end

    # @param [String] ending
    #
    # @raise [ArgumentError]
    #
    # @return [String]
    def ensure_ending_char!(ending)
      Ruuuby::Params::check_char(ending)
      self << ending if self.length == 0 || self[-1] != ending
      self
    end
  end

end
